/*
 * Entagged Audio Tag library
 * Copyright (c) 2003-2005 Raphal Slinckx <raphael@slinckx.net>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *  
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package entagged.audioformats.mp3.util;

import entagged.audioformats.exceptions.*;
import entagged.audioformats.mp3.*;

import java.io.*;
import java.nio.*;

public class Id3v2TagReader {
	
	boolean[] ID3Flags;
	Id3v2TagSynchronizer synchronizer = new Id3v2TagSynchronizer();
	
	Id3v23TagReader v23 = new Id3v23TagReader();
	
	public Id3v2Tag read(RandomAccessFile raf) throws CannotReadException, IOException {
		Id3v2Tag tag;

		byte[] b = new byte[3];

		raf.read(b);
		raf.seek(0);

		String ID3 = new String(b);

		if (!ID3.equals("ID3")) {
			throw new CannotReadException("Not an ID3 tag");
		}
		//Begins tag parsing ---------------------------------------------
		raf.seek(3);
		//----------------------------------------------------------------------------
		//Version du tag ID3v2.xx.xx
		String versionHigh=raf.read() +"";
		String versionID3 =versionHigh+ "." + raf.read();
		//System.err.println(f+" Version:"+versionID3);
		//if (!versionHigh.equals("3"))
		//		throw new CannotReadException("ID3v2 tag version "+ versionID3 + " not supported yet");
		//------------------------------------------------------------------------- ---
		//D?tection de certains flags (A COMPLETER)
		ID3Flags = processID3Flags(raf.readByte());
		//----------------------------------------------------------------------------
		
//			On extrait la taille du tag ID3
		int tagSize = readSyncsafeInteger(raf);
		//System.err.println("TagSize: "+tagSize);
		
//			------------------NEWNEWNWENENEWNENWEWN-------------------------------
		//Fill a byte buffer, then process according to correct version
		b = new byte[tagSize+2];
		raf.readFully(b);
		ByteBuffer bb = ByteBuffer.wrap(b);
		
		if (ID3Flags[0]==true) {
		    //We have unsynchronization, first re-synchronize
		    bb = synchronizer.synchronize(bb);
		}
		
		if(versionHigh.equals("2")) {
			tag = v23.read(bb, ID3Flags, Id3v2Tag.ID3V22);
		}
		else if(versionHigh.equals("3")) {
		    tag = v23.read(bb, ID3Flags, Id3v2Tag.ID3V23);
		}
		else if(versionHigh.equals("4")) {
		    //tag = v24.read(raf, ID3Flags);
			throw new CannotReadException("ID3v2 tag version "+ versionID3 + " not supported !");
		}
		else {
			throw new CannotReadException("ID3v2 tag version "+ versionID3 + " not supported !");
		}
		
		return tag;
	}
	
	/**
	 *  Converts a byte representing a series of Flags to an array of booleans for an easier handling
	 *
	 * @param  b  A byte containing ID3 Flags
	 * @return    An array of booleans reflecting the state of each bit in the byte representing the ID3 Flags
	 */
	private boolean[] processID3Flags(byte b) {
		boolean[] flags;

		if (b != 0) {
			flags = new boolean[4];

			int flag = b & 128;

			if (flag == 128)
				flags[0] = true;
			else
				flags[0] = false; //unsynchronisation
			flag = b & 64;
			if (flag == 64)
				flags[1] = true;
			else
				flags[1] = false; //Extended Header
			flag = b & 32;
			if (flag == 32)
				flags[2] = true;
			else
				flags[2] = false; //Experimental Indicator
			flag = b & 16;
			if (flag == 16)
				flags[3] = true;
			else
				flags[3] = false; //Footer Present
		} else {
			flags = new boolean[4];
			flags[0] = false;
			flags[1] = false;
			flags[2] = false;
			flags[3] = false;
		}
		//System.err.println("Flags:unsynchronisation "+flags[0]+"Extended Header"+flags[1]);
		return flags;
	}
	
	/**
	 *  Take the first four bytes from the provided MBB and convert them into an int, by treating them as a sync. safe integer.
	 *
	 * @param  raf  A MappedByteBuffer containing the bytes to be dealt with.
	 * @return      The integer value represented by the bytes.
	 * @exception IOException IO Error
	 */
	private int readSyncsafeInteger(RandomAccessFile raf)	throws IOException {
		int value = 0;

		value += (raf.read()& 0xFF) << 21;
		value += (raf.read()& 0xFF) << 14;
		value += (raf.read()& 0xFF) << 7;
		value += raf.read()& 0xFF;

		return value;
	}
}
